@@ -1,17 +1,19 @@
 #!/usr/bin/env python
 __author__ = ('Aitor Blanco-Miguez (aitor.blancomiguez@unitn.it), '
               'Francesco Beghini (francesco.beghini@unitn.it), '
+              'Moreno Zolfo (moreno.zolfo@unitn.it), '
               'Nicola Segata (nicola.segata@unitn.it), '
               'Duy Tin Truong, '
-              'Francesco Asnicar (f.asnicar@unitn.it)')
-__version__ = '4.0.6'
-__date__ = '1 Mar 2023'
+              'Francesco Asnicar (f.asnicar@unitn.it), ' 
+              'Claudia Mengoni (claudia.mengoni@unitn.it)')
+__version__ = '4.1.2'
+__date__ = '16 Oct 2025'
 
 import sys
 try:
     from metaphlan import mybytes, plain_read_and_split, plain_read_and_split_line, read_and_split, read_and_split_line, check_and_install_database, remove_prefix
 except ImportError:
-    sys.exit("CRITICAL ERROR: Unable to find the MetaPhlAn python package. Please check your install.") 
+    sys.exit("CRITICAL ERROR: Unable to find the MetaPhlAn python package. Please check your install.")
 
 if float(sys.version_info[0]) < 3.0:
     sys.stderr.write("MetaPhlAn requires Python 3, your current Python version is {}.{}.{}\n"
@@ -28,14 +30,25 @@ from glob import glob
 from subprocess import DEVNULL
 import argparse as ap
 import bz2
+import gzip
 import pickle
 import subprocess as subp
 import tempfile as tf
 
+from Bio.Seq import Seq
+from Bio import SeqIO
+from Bio.SeqRecord import SeqRecord
+from collections import Counter
 try:
-    import numpy as np
+    from .utils.parallelisation import execute_pool
 except ImportError:
-    sys.stderr.write("Error! numpy python library not detected!!\n")
+    from utils.parallelisation import execute_pool
+try:
+    import pandas as pd
+    import numpy as np
+    import pysam
+except Exception as e:
+    sys.stderr.write("Error! python library not detected\n{}\n".format(e))
     sys.exit(1)
 
 #**********************************************************************************************
@@ -69,7 +82,7 @@ INDEX = 'latest'
 tax_units = "kpcofgst"
 
 def read_params(args):
-    p = ap.ArgumentParser( description=
+    p = ap.ArgumentParser( description =
             "DESCRIPTION\n"
             " MetaPhlAn version "+__version__+" ("+__date__+"): \n"
             " METAgenomic PHyLogenetic ANalysis for metagenomic taxonomic profiling.\n\n"
@@ -297,6 +310,13 @@ def read_params(args):
     arg( '--min_ab', metavar="", default=0.1, type=float, help =
          "The minimum percentage abundance for the clade in the clade_specific_strain_tracker analysis\n"  )
 
+    g = p.add_argument_group('Viral Sequence Clusters Analisys')
+    arg = g.add_argument
+    arg("--profile_vsc", action="store_true",help="Add this parameter to profile Viruses with VSCs approach.")
+    arg("--vsc_out", help="Path to the VSCs breadth-of-coverage output file", default="mp3_viruses.csv")
+    arg("--vsc_breadth", help="Minimum Breadth of Coverage for a Viral Group to be reported.\n"
+        "Default is 0.75 (at least 75 percent breadth to report)", default=0.75,type=float)
+
     g = p.add_argument_group('Output arguments')
     arg = g.add_argument
     arg( '-o', '--output_file',  metavar="output file", type=str, default=None, help =
@@ -335,8 +355,18 @@ def read_params(args):
         help="The number of CPUs to use for parallelizing the mapping [default 4]")
     arg('--subsampling', type=int, default=None,
         help="Specify the number of reads to be considered from the input metagenomes [default None]")
+    arg('--subsampling_output', type=str, default=None,
+        help="The output file for the subsampled reads. If --subsampling_paired is specified two files are created with suffixes R1 and R2. If not specified the subsampled reads will not be saved.")
+    arg('--subsampling_paired',  type=int, default=None,
+        help="Specify the number of paired reads to be considered from the input metagenomes [default None]")
+    arg('-1', type=str, default=None, metavar='FORWARD_READS',
+        help="Specify the fastq file with forward reads of the input metagenomes. Reads are assumed to be in the same order in the forward and reverse files! [default None]")
+    arg('-2', type=str, default=None, metavar='REVERSE_READS',
+        help="Specify the fastq file with reverse reads of the input metagenomes. Reads are assumed to be in the same order in the forward and reverse files! [default None]")
+    arg('--mapping_subsampling', action='store_true',
+        help="If used, the subsamping will be done on the mapping results instead of on the reads.")
     arg('--subsampling_seed', type=str, default='1992',
-        help="Random seed to use in the selection of the subsampled reads. Choose \"random\r for a random behaviour")
+        help="Random seed to use in the selection of the subsampled reads. Choose \"random\r for a random behaviour") 
     arg('--install', action='store_true',
         help="Only checks if the MetaPhlAn DB is installed and installs it if not. All other parameters are ignored.")
     arg('--offline', action='store_true',
@@ -346,13 +376,50 @@ def read_params(args):
     arg('--read_min_len', type=int, default=70,
         help="Specify the minimum length of the reads to be considered when parsing the input file with "
              "'read_fastx.py' script, default value is 70")
-    arg('-v', '--version', action='version',
-        version="MetaPhlAn version {} ({})".format(__version__, __date__),
+    arg('-v', '--version', action='store_true',
         help="Prints the current MetaPhlAn version and exit")
     arg("-h", "--help", action="help", help="show this help message and exit")
 
     return vars(p.parse_args())
 
+def get_installed_db_version(db_dir: str = None):
+    """
+    Return a sorted list of DB base names (e.g., mpa_vJan24_CHOCOPhlAnSGB_202401)
+    for which both .pkl and .fna exist in db_dir.
+    """
+    if db_dir:
+        db_folder = db_dir
+    else:
+        db_folder = DEFAULT_DB_FOLDER
+    if not os.path.exists(db_folder):
+        return None
+    files = os.listdir(db_folder)
+    pkl_bases = {os.path.splitext(f)[0] for f in files if f.endswith(".pkl")}
+    bt2l_bases = {os.path.basename(f).split('.')[0] for f in files if f.endswith(".bt2l")}
+    common = sorted(pkl_bases & bt2l_bases)
+    return common or None
+
+def get_cli_db_dir(argv=None):
+    """
+    Robustly detect database folder from CLI (--db_dir or --bowtie2db),
+    supporting both '--flag value' and '--flag=value'.
+    """
+    if argv is None:
+        argv = sys.argv[1:]
+
+    p = ap.ArgumentParser(add_help=False)
+    p.add_argument("--db_dir", dest="db_dir")
+    p.add_argument("--bowtie2db", dest="bowtie2db")
+    # Ignore everything else for now
+    args, _ = p.parse_known_args(argv)
+
+    val = args.db_dir or args.bowtie2db
+    if not val:
+        return None
+    # normalize: expand ~ and $VARS, strip trailing slash, absolutize
+    val = os.path.abspath(os.path.expanduser(os.path.expandvars(val).rstrip("/")))
+    return val
+
 def set_mapping_arguments(index, bowtie2_db):
     mpa_pkl = 'mpa_pkl'
     bowtie2db = 'bowtie2db'
@@ -366,8 +433,125 @@ def set_mapping_arguments(index, bowtie2_db):
 
     return (mpa_pkl, bowtie2db)
 
+
+def set_vsc_parameters(index, bowtie2_db):
+    vsc_fna = os.path.join(bowtie2_db, "{}_VSG.fna".format(index))
+    vsc_vinfo = os.path.join(bowtie2_db, "{}_VINFO.csv".format(index))
+
+    if not os.path.isfile(vsc_fna):
+        sys.stderr.write("Error:\n {} file not found in bowtie2db folder ({}). Re-download MetaPhlAn database.".format("{}_VSG.fna".format(index, bowtie2_db)))
+        sys.exit(1)
+
+    if not os.path.isfile(vsc_vinfo):
+        sys.stderr.write("Error:\n {} file not found in bowtie2db folder ({}). Re-download MetaPhlAn database.".format("{}_VINFO.csv".format(index, bowtie2_db)))
+        sys.exit(1)
+
+    return(vsc_fna, vsc_vinfo)
+
+def vsc_bowtie2(profile_vsc_folder, nproc, file_format="fasta",
+                exe=None,bt2build_exe=None, min_alignment_len=None, read_min_len=0):
+
+    try:
+        subp.check_call([exe if exe else 'bowtie2', "-h"], stdout=DEVNULL)
+        subp.check_call([bt2build_exe if bt2build_exe else 'bowtie2-build', "-h"], stdout=DEVNULL)
+    except Exception as e:
+        sys.stderr.write('OSError: "{}"\nFatal error running BowTie2 and bowtie2-build. Is BowTie2 in the system path?\n'.format(e))
+        sys.exit(1)
+
+    # checking files
+    markerfile= profile_vsc_folder+'/v_mks.fa' 
+    readsfile= profile_vsc_folder+'/v_reads.fq' 
+    vscBamFile= profile_vsc_folder+'/v_align.bam'
+
+    if not os.path.exists(markerfile) or not os.path.exists(readsfile):
+        
+        sys.stderr.write('Error:\nThere was an error in the VSCs file lookup.\n \
+            It may be that there are not enough reads to profile (not enough depth).\n \
+            Passing without reporting any virus.\n')
+        return []
+
+    if os.stat(markerfile).st_size == 0:
+        sys.stderr.write('Warning:\nNo reads aligning to VSC markers in this file.\n')        
+        #return an empty list. MetaPhlAn will continue as normal, without VSCs
+        return []
+
+    #build the db
+
+    bt2build_call = bt2build_exe if bt2build_exe else 'bowtie2-build'
+    dbpath = profile_vsc_folder+'/v_mks'
+    subp.check_call( [bt2build_call, markerfile, dbpath, '-q','--threads', str(nproc)] )
+
+
+    if not all([os.path.exists(".".join([str(dbpath), p]))
+                            for p in ["1.bt2", "2.bt2", "3.bt2", "4.bt2", "rev.1.bt2", "rev.2.bt2"]]):
+        sys.stderr.write('Error:\nThere was an error in running bowtie2-map and not all files were generated. Stopping.\n')
+        sys.exit(1)
+
+    try:
+
+        bt2_call = exe if exe else 'bowtie2'
+        bt2_command = [bt2_call,'--very-sensitive','--quiet','-p',str(nproc),'-x',dbpath,'--no-unal','-U',readsfile]
+        stv_command = ['samtools','view','-bS','-','-@',str(nproc)]
+        sts_command = ['samtools','sort','-','-@',str(nproc),'-o',vscBamFile]
+        sti_command = ['samtools','index',vscBamFile]
+
+        p3 = subp.Popen(sts_command, stdin=subp.PIPE)
+        p2 = subp.Popen(stv_command, stdin=subp.PIPE, stdout=p3.stdin)
+        p1 = subp.Popen(bt2_command, stdout=p2.stdin)
+
+        p1.communicate() 
+        p2.communicate()
+        p3.communicate()
+
+        #index
+        subp.check_call(sti_command)
+
+    except Exception as e:
+        sys.stderr.write('Error: "{}"\nFatal error running BowTie2 for Viruses\n'.format(e))
+        sys.exit(1)
+
+    if not os.path.exists(vscBamFile):
+        sys.stderr.write('Error:\nUnable to create BAM FILE for viruses\n')
+        sys.exit(1)
+
+    try:
+        bamHandle = pysam.AlignmentFile(vscBamFile, "rb")
+    except Exception as e:
+        sys.stderr.write('Error: "{}"\nCheck PySam is correctly working\n'.format(e))
+        sys.exit(1)
+
+    VSC_report=[]
+
+    for c, length in zip(bamHandle.references,bamHandle.lengths):
+        coverage_positions = {}
+
+        for pileupcolumn in bamHandle.pileup(c):
+
+            tCoverage = 0
+            for pileupread in pileupcolumn.pileups:
+
+                if not pileupread.is_del and not pileupread.is_refskip \
+                        and pileupread.alignment.query_qualities[pileupread.query_position] >= 20 \
+                        and pileupread.alignment.query_sequence[pileupread.query_position].upper() in ('A','T','C','G'):
+                        tCoverage +=1
+
+            if tCoverage >= 1:
+                coverage_positions[pileupcolumn.pos] = tCoverage
+        
+        breadth = float(len(coverage_positions.keys()))/float(length)
+        
+        if breadth > 0:
+            cvals=list(coverage_positions.values())
+            VSC_report.append({'M-Group/Cluster':c.split('|')[2].split('-')[0], 'genomeName':c, 'len':length, 'breadth_of_coverage':breadth, 'depth_of_coverage_mean': np.mean(cvals), 'depth_of_coverage_median': np.median(cvals)})
+
+    if bamHandle:
+        bamHandle.close()
+
+    return VSC_report
+
+
 def run_bowtie2(fna_in, outfmt6_out, bowtie2_db, preset, nproc, min_mapq_val, file_format="fasta",
-                exe=None, samout=None, min_alignment_len=None, read_min_len=0):
+                exe=None, samout=None, min_alignment_len=None, read_min_len=0, profile_vsc_folder=False):
     # checking read_fastx.py
     read_fastx = "read_fastx.py"
 
@@ -409,6 +593,11 @@ def run_bowtie2(fna_in, outfmt6_out, bowtie2_db, preset, nproc, min_mapq_val, fi
         p = subp.Popen(bowtie2_cmd, stdout=subp.PIPE, stdin=readin.stdout)
         readin.stdout.close()
         lmybytes, outf = (mybytes, bz2.BZ2File(outfmt6_out, "w")) if outfmt6_out.endswith(".bz2") else (str, open(outfmt6_out, "w"))
+
+        if profile_vsc_folder:
+            CREAD=[]
+            list_of_viral_markers = open(profile_vsc_folder+'/viralmk.txt','w')
+
         try:
             if samout:
                 if samout[-4:] == '.bz2':
@@ -429,8 +618,28 @@ def run_bowtie2(fna_in, outfmt6_out, bowtie2_db, preset, nproc, min_mapq_val, fi
                         if mapq_filter(o[2], int(o[4]), min_mapq_val) :  # filter low mapq reads
                             if ((min_alignment_len is None) or
                                     (max([int(x.strip('M')) for x in re.findall(r'(\d*M)', o[5]) if x]) >= min_alignment_len)):
+                                                                # Profile viral markers in a different way
+                                if profile_vsc_folder and o[2].startswith('VDB|'):
+
+                                    mCluster = o[2]
+                                    mGroup = o[2].split('|')[2].split('-')[0]
+
+                                    list_of_viral_markers.write(mGroup+'\t'+mCluster+'\n')
+
+                                    if (hex(int(o[1]) & 0x10) == '0x0'): #front read
+                                        rr=SeqRecord(Seq(o[9]),letter_annotations={'phred_quality':[ord(_)-33 for _ in o[10][::-1]]}, id=o[0])
+                                    else:
+                                        rr=SeqRecord(Seq(o[9]).reverse_complement(),letter_annotations={'phred_quality':[ord(_)-33 for _ in o[10][::-1]]}, id=o[0])
+
+                                    CREAD.append(rr)
+
+                                # normal route for non-viral markers
                                 outf.write(lmybytes("\t".join([ o[0], o[2].split('/')[0] ]) + "\n"))
 
+        if profile_vsc_folder and os.path.isdir(profile_vsc_folder):
+            SeqIO.write(CREAD,profile_vsc_folder+'/v_reads.fq','fastq')
+            list_of_viral_markers.close()
+
         if samout:  
             sam_file.close()
 
@@ -814,7 +1023,8 @@ class TaxTree:
         return ret_d, ret_r, tot_reads
 
 def mapq_filter(marker_name, mapq_value, min_mapq_val):
-    if 'GeneID:' in marker_name:
+    ##if 'GeneID:' in marker_name:
+    if 'GeneID:' in marker_name or 'VDB' in marker_name:
         return True
     else:
         if mapq_value > min_mapq_val:
@@ -828,7 +1038,7 @@ def separate_reads2markers(reads2markers):
     else:
         return {r: m for r, m in reads2markers.items() if ('SGB' in m or 'EUK' in m) and not 'VDB' in m}, {r: m for r, m in reads2markers.items() if 'VDB' in m and not ('SGB' in m or 'EUK' in m)}
 
-def map2bbh(mapping_f, min_mapq_val, input_type='bowtie2out', min_alignment_len=None, nreads=None, subsampling=None, subsampling_seed='1992'):
+def map2bbh(mapping_f, min_mapq_val, input_type='bowtie2out', min_alignment_len=None, nreads=None, mapping_subsampling=False, subsampling=None, subsampling_seed='1992', remove_input=False):
     if not mapping_f:
         ras, ras_line, inpf = plain_read_and_split, plain_read_and_split_line, sys.stdin
     else:
@@ -862,11 +1072,9 @@ def map2bbh(mapping_f, min_mapq_val, input_type='bowtie2out', min_alignment_len=
                     reads2markers[o[0]] = o[2].split('/')[0]
     inpf.close()
     
-    if subsampling != None:
+    if subsampling is not None and mapping_subsampling:
         if subsampling >= n_metagenome_reads:
-            sys.stderr.write("WARNING: The specified subsampling ({}) is higher than the original number of reads ({}).".format(subsampling, n_metagenome_reads))
-        elif subsampling < 10000:
-            sys.stderr.write("WARNING: The specified subsampling ({}) is below the recommended minimum of 10,000 reads.".format(subsampling))
+            sys.stderr.write("WARNING: The specified subsampling ({}) is equal or higher than the original number of reads ({}). Subsampling will be skipped.\n".format(subsampling, n_metagenome_reads))
         else:
             reads2markers =  dict(sorted(reads2markers.items()))
             if subsampling_seed.lower() != 'random':
@@ -882,8 +1090,8 @@ def map2bbh(mapping_f, min_mapq_val, input_type='bowtie2out', min_alignment_len=
             sgb_reads2markers.clear()
             viral_reads2markers.clear()
             n_metagenome_reads = subsampling
-    elif n_metagenome_reads < 10000:
-        sys.stderr.write("WARNING: The number of reads in the sample ({}) is below the recommended minimum of 10,000 reads.".format(subsampling))
+    elif subsampling is None and n_metagenome_reads < 10000:
+        sys.stderr.write("WARNING: The number of reads in the sample ({}) is below the recommended minimum of 10,000 reads.\n".format(n_metagenome_reads))
         
     markers2reads = defdict(set)   
     for r, m in reads2markers.items():
@@ -976,7 +1184,111 @@ def maybe_generate_biom_file(tree, pars, abundance_predictions):
 
     return True
 
+def _make_gen_fastq(reader):
+    b = reader(1024 * 1024) 
+    while (b):
+        yield b
+        b = reader(1024 * 1024)
+
+def rawpycount(filename):
+    f = bz2.BZ2File(filename, 'rb') if filename.endswith(".bz2") else gzip.open(filename, 'rb') if filename.endswith('.gz') else open(filename, 'rb')
+    f_gen = _make_gen_fastq(f.read)
+    n_metagenome_reads = sum( buf.count(b'\n') for buf in f_gen)
+    f.close()
+    return n_metagenome_reads
+
+def subsample_file(inp, paired, subsampling, out, sample):         
+    length, read_num = 0, -1
+    counter=0
+    out_l = out
+    for inp_f in inp.split(','):
+        if paired:
+            length, read_num = 0, -1  
+            out=out_l[counter]
+            counter+=1
+        line = 1
+        reader = bz2.open(inp_f, 'rt') if inp_f.endswith(".bz2") else gzip.open(inp_f, 'rt') if inp_f.endswith('.gz') else open(inp_f, 'r')
+        while (line and length < subsampling):
+            line = reader.readline()
+            read_num += 1
+            if read_num in sample:
+                length += 1
+                out.write(line)       
+                for _ in range(3):
+                    out.write(reader.readline())
+            else:
+                for _ in range(3):
+                    reader.readline()
+        reader.close()
+    out.close()
+    
+def subsample_reads(inp, subsampling, subsampling_seed, subsampling_output, tmp_dir, paired):
+
+    n_metagenome_reads = execute_pool(((rawpycount, inp_f) for inp_f in inp.split(',')), 2)
+    n_metagenome_reads = [int(n/4) for n in n_metagenome_reads]
+
+    if paired:
+        subsampling //= 2
+        if n_metagenome_reads[0] != n_metagenome_reads[1]:
+            sys.stderr.write("Error: The specified reads file are not the same length! Make sure the forward and reverse reads are files are not damaged and reads are in the same order. Exiting ...\n\n") 
+            sys.exit(1)
+        n_metagenome_reads = n_metagenome_reads[0]
+    else:
+        n_metagenome_reads = sum(n_metagenome_reads)
+
+    if subsampling >= n_metagenome_reads:
+        sys.stderr.write("WARNING: The specified subsampling ({}) is equal or higher than the original number of reads ({}). Subsampling will be skipped.\n".format(subsampling, sum(n_metagenome_reads))) 
+        return inp, None
+
+    if subsampling_seed.lower() != 'random':
+        random.seed(int(subsampling_seed))
+        
+    sample = set(random.sample(range(n_metagenome_reads), subsampling))
+    
+    if subsampling_output is None:
+        if paired:
+            subsampling_output = list()
+            out=list()
+            for _ in inp.split(','):
+                out_f = tf.NamedTemporaryFile(dir=tmp_dir, mode='w', delete=False)
+                out.append(out_f)
+                subsampling_output.append(out_f.name)
+        else:
+            out = tf.NamedTemporaryFile(dir=tmp_dir, mode='w', delete=False)
+            subsampling_output=out.name
+    else:
+        if paired:
+            r, ext = os.path.splitext(subsampling_output)
+
+            subsampling_output = list()
+            out = list()
+
+            subsampling_output.append('.'.join([r, 'R1'+ext]))
+            subsampling_output.append('.'.join([r, 'R2'+ext]))
+
+            for s in subsampling_output:
+                out.append(bz2.open(s, 'wt') if s.endswith(".bz2") else gzip.open(s, 'wt') if s.endswith('.gz') else open(s, 'w'))
+        else:
+            out = bz2.open(subsampling_output, 'wt') if subsampling_output.endswith(".bz2") else gzip.open(subsampling_output, 'wt') if subsampling_output.endswith('.gz') else open(subsampling_output, 'w')
+            
+    subsample_file(inp, paired, subsampling, out, sample)
+
+    if isinstance(subsampling_output, list):
+        subsampling_output = ','.join(subsampling_output)
+        
+    return subsampling_output, subsampling
+
 def main():
+    if '-v' in sys.argv or '--version' in sys.argv:
+        print(f"MetaPhlAn version {__version__} ({__date__})")
+        db_dir=get_cli_db_dir()
+        db_version=get_installed_db_version(db_dir)
+        if db_version:
+            print("Installed databases:", ", ".join(db_version))
+        else:
+            print("No complete MetaPhlAn Bowtie2 database found")
+        exit(0)
+
     ranks2code = { 'k' : 'superkingdom', 'p' : 'phylum', 'c':'class',
                    'o' : 'order', 'f' : 'family', 'g' : 'genus', 's' : 'species'}
     pars = read_params(sys.argv)
@@ -986,12 +1298,54 @@ def main():
     SGB_ANALYSIS = not pars['mpa3']
 
     ESTIMATE_UNK = pars['unclassified_estimation']
+
+    if pars['subsampling'] and pars['subsampling_paired']:
+        sys.stderr.write("Error: You specified both --subsampling and --subsampling_paired. Choose only one of the two options. Exiting...")
+        sys.exit(1)
+
+    if pars['subsampling']:
+        sys.stderr.write("WARNING: If you use --subsampling the reads will be subsampled NOT taking into account paired information. Use --subsampling_paired if you have paired ends reads.\n".format(pars['subsampling']))
+    
+    if pars['mapping_subsampling'] and pars['subsampling'] is None:
+        sys.stderr.write("Error: The --mapping_subsampling parameter should be used together with the --subsampling parameter. Exiting...\n\n")
+        sys.exit(1)
+
+    if pars['subsampling_paired']:
+        subsampling_paired=True
+        pars['subsampling']=pars['subsampling_paired']
+    else:
+        subsampling_paired=False
+
+    if subsampling_paired and (pars['1'] is None or pars['2'] is None):
+        sys.stderr.write("Error: If you specify --subsampling_paired you have to provide forward and reverse reads as -1 and -2 respectively. Reads are assumed to be in the same order in the two files.\n Exiting...\n\n".format(pars['subsampling']))
+        sys.exit(1)
+    
+    if pars['subsampling'] is not None and pars['subsampling'] < 10000:
+        sys.stderr.write("WARNING: The specified subsampling ({}) is below the recommended minimum of 10,000 reads.\n".format(pars['subsampling']))
     
     if not (pars['subsampling_seed'].lower() == 'random' or pars['subsampling_seed'].isdigit()):
         sys.stderr.write("Error: The --subsampling_seed parameter is not accepted. It should contain an integer number or \"random\". Exiting...\n\n")
         sys.exit(1)
-    
+        
+    if not pars['mapping_subsampling'] and pars['subsampling'] is not None:
+        if pars['input_type'] != 'fastq':
+            sys.stderr.write("Error: The reads' subsampling procedure requires fastq input! Exiting...\n\n")
+            sys.exit(1)
+        elif (not pars['inp']) and (not subsampling_paired):
+            sys.stderr.write("Error: Input reads for the subsampling must be provided as parameter. Stdin input is not allowed! Exiting...\n\n")
+            sys.exit(1)
+        
+        if subsampling_paired:
+            if not os.path.exists(pars['2']) or not os.path.exists(pars['1']):
+                sys.stderr.write("Error: Files passed with -1 ({}) or -2 ({}) not found. Exiting...\n\n".format(pars['1'],pars['2']))
+                sys.exit(1)
 
+            if pars['inp']:
+                sys.stderr.write("WARNING: since --subsampling_paired has been specified, reads are taken from -1 ({}) and -2 ({}), not from -inp.\n".format(pars['1'],pars['2']))
+            pars['inp'] = pars['1']+','+pars['2']
+
+        pars['inp'], pars['subsampling'] = subsample_reads(pars['inp'], pars['subsampling'], pars['subsampling_seed'], pars['subsampling_output'], pars['tmp_dir'], paired=subsampling_paired)
+        
     # check if the database is installed, if not then install
     pars['index'] = check_and_install_database(pars['index'], pars['bowtie2db'], pars['bowtie2_build'], pars['nproc'], pars['force_download'], pars['offline'])
 
@@ -999,6 +1353,37 @@ def main():
         sys.stderr.write('The database is installed\n')
         return
 
+    #if we are to profile viral clusters:
+
+    if pars['profile_vsc']:
+        #TODO: comply to other input types
+        if pars['input_type'] != 'fastq':
+            sys.stderr.write("The Viral Sequence Clusters mode requires fastq input!\n")
+            sys.exit(1)
+        
+        if not pars['vsc_out']:
+            sys.stderr.write("The Viral Sequence Clusters mode requires to specify an output file with the --vsc_out parameter\n")
+            sys.exit(1)
+
+        tmpvirdir=tf.TemporaryDirectory(dir=pars['tmp_dir'])
+        viralTempFolder=tmpvirdir.name
+        
+        # Uncomment to preserve tmp dir
+        #tmpvirdir=tf.mkdtemp(dir=pars['tmp_dir'])
+        #viralTempFolder=tmpvirdir
+        #print("VF: ",viralTempFolder)
+
+        vsc_fna, vsc_vinfo = set_vsc_parameters(pars['index'], pars['bowtie2db'])
+
+        # implies SGBs (uses the same database with SGBs, so this is needed to avoid
+        # the need to specify --sgb when using --profile_vscs)
+        SGB_ANALYSIS = True
+
+
+    else:
+        viralTempFolder = None
+
+
     # set correct map_pkl and bowtie2db variables
     pars['mpa_pkl'], pars['bowtie2db'] = set_mapping_arguments(pars['index'], pars['bowtie2db'])
 
@@ -1074,7 +1459,11 @@ def main():
             run_bowtie2(pars['inp'], pars['bowtie2out'], pars['bowtie2db'],
                                 pars['bt2_ps'], pars['nproc'], file_format=pars['input_type'],
                                 exe=pars['bowtie2_exe'], samout=pars['samout'],
-                                min_alignment_len=pars['min_alignment_len'], read_min_len=pars['read_min_len'], min_mapq_val=pars['min_mapq_val'])
+
+                                min_alignment_len=pars['min_alignment_len'], read_min_len=pars['read_min_len'], min_mapq_val=pars['min_mapq_val'],profile_vsc_folder=viralTempFolder)
+            if pars['subsampling_output'] is None and not pars['mapping_subsampling'] and pars['subsampling'] is not None:
+                for inp_f in pars['inp'].split(','):
+                    os.remove(inp_f)
             pars['input_type'] = 'bowtie2out'
         pars['inp'] = pars['bowtie2out'] # !!!
     with bz2.BZ2File( pars['mpa_pkl'], 'r' ) as a:
@@ -1091,7 +1480,57 @@ def main():
                 "\nExiting...\n\n" )
         sys.exit(1)
 
-    markers2reads, n_metagenome_reads, avg_read_length = map2bbh(pars['inp'], pars['min_mapq_val'], pars['input_type'], pars['min_alignment_len'], pars['nreads'], pars['subsampling'], pars['subsampling_seed'])
+    markers2reads, n_metagenome_reads, avg_read_length = map2bbh(pars['inp'], pars['min_mapq_val'], pars['input_type'], pars['min_alignment_len'], pars['nreads'], pars['mapping_subsampling'], pars['subsampling'], pars['subsampling_seed'])
+
+    if pars['profile_vsc']:
+        
+        try:
+            VSCs_markers = SeqIO.index(vsc_fna, "fasta")
+        except Exception as e:
+            sys.stderr.write('Error: "{}"\nCould not access to {}.\n'.format(e,vsc_fna))
+            sys.exit(1)
+
+        with open(viralTempFolder+'/viralmk.txt') as viral_markers_to_remap:
+
+            allViralMarkers = {}
+            for vmarker in viral_markers_to_remap:
+                group,marker = vmarker.strip().split()
+                if group not in allViralMarkers:
+                    allViralMarkers[group] = [marker]
+                else:
+                    allViralMarkers[group].append(marker)
+
+            selectedMakrers=[]
+            for grp,v in allViralMarkers.items():
+
+                cv=Counter(v)
+                if (len(cv) > 1):
+                    normalized_occurrencies=sorted([(mk,mk_occurrencies,len(VSCs_markers[mk].seq)) for (mk,mk_occurrencies) in cv.items()],key=lambda x: x[1]/x[2],reverse=True)
+                    topMarker= VSCs_markers[normalized_occurrencies[0][0]] #name of the viralGenome
+                else:
+                    topMarker=VSCs_markers[v[0]] # the only hitted viralGenome is the winning one
+
+                selectedMakrers.append(topMarker)
+
+
+            SeqIO.write( selectedMakrers,viralTempFolder+'/v_mks.fa','fasta' )
+
+            VSC_report = vsc_bowtie2(viralTempFolder, pars['nproc'], file_format=pars['input_type'],
+                        exe=pars['bowtie2_exe'], bt2build_exe=pars['bowtie2_build']) 
+
+
+            if not VSC_report == []:
+                with open(pars['vsc_out'],'w') as outf:
+
+                    outf.write('#{}\n'.format(pars['index']))
+                    outf.write('#{}\n'.format(' '.join(sys.argv)))
+                    outf.write('#SampleID\t{}\n'.format(pars['sample_id']))
+                    vsc_out_df = pd.DataFrame.from_dict(VSC_report).query('breadth_of_coverage >= {}'.format(pars['vsc_breadth']))
+                    vsc_info_df = pd.read_table(vsc_vinfo, sep='\t')
+                    vsc_out_df = vsc_out_df.merge(vsc_info_df, on='M-Group/Cluster').sort_values(by='breadth_of_coverage', ascending=False).set_index('M-Group/Cluster')
+
+                    vsc_out_df.to_csv(outf,sep='\t',na_rep='-')
+
 
     tree.set_stat( pars['stat'], pars['stat_q'], pars['perc_nonzero'], avg_read_length, pars['avoid_disqm'])
 
@@ -1125,6 +1564,9 @@ def main():
             outf.write('#{}\n'.format(pars['index']))
             outf.write('#{}\n'.format(' '.join(sys.argv)))
             outf.write('#{} reads processed\n'.format(n_metagenome_reads))
+        
+        if pars['t'] == 'rel_ab_w_read_stats':
+            outf.write('#Average read length {}\n'.format(avg_read_length))           
 
         if not CAMI_OUTPUT:
             outf.write('#' + '\t'.join((pars["sample_id_key"], pars["sample_id"])) + '\n')
@@ -1293,9 +1735,7 @@ def main():
                 sys.stderr.write("Clade "+pars['clade']+" not present at an abundance >"+str(round(pars['min_ab'],2))+"%, "
                                  "so no clade specific markers are reported\n")
 
-
 if __name__ == '__main__':
     t0 = time.time()
     main()
-    sys.stderr.write('Elapsed time to run MetaPhlAn: {} s\n'.format( (time.time()-t0) ) )
-
+    sys.stderr.write('Elapsed time to run MetaPhlAn: {} s\n'.format( (time.time()-t0) ) )
\ No newline at end of file
