@@ -1,11 +1,13 @@
 __author__ = 'Aitor Blanco Miguez (aitor.blancomiguez@unitn.it)'
-__version__ = '4.0.6'
-__date__ = '1 Mar 2023'
+__version__ = '4.1.2'
+__date__ = '16 Oct 2025'
 
 
 import os
 import pickle
 import bz2
+
+import pandas as pd
 from Bio import SeqIO
 try:
     from .external_exec import generate_markers_fasta
@@ -27,6 +29,7 @@ class MetaphlanDatabaseController:
             if verbose:
                 info('Done.')
 
+
     def get_database_name(self):
         """Gets database name
 
@@ -35,14 +38,46 @@ class MetaphlanDatabaseController:
         """
         return self.database.split('/')[-1][:-4]
 
-    def get_markers2species(self):
+
+    def get_markers2clade(self):
         """Retrieve information from the MetaPhlAn database
 
         Returns:
-            dict: the dictionary assigning markers to clades
+            dict[str, str]: the dictionary assigning markers to clades
         """
         self.load_database()
-        return {marker: self.database_pkl['markers'][marker]['clade'] for marker in self.database_pkl['markers']}
+        return {marker_name: marker_info['clade'] for marker_name, marker_info in self.database_pkl['markers'].items()}
+
+
+    def get_clade2markers(self):
+        markers2clade = self.get_markers2clade()
+        markers2clade = pd.Series(markers2clade)
+        return markers2clade.groupby(markers2clade).groups
+
+
+    def get_markers_for_clade(self, clade):
+        """
+
+        Args:
+            clade (str):
+
+        Returns:
+            set: marker names for the given clade
+
+        """
+        self.load_database()
+        return set(marker_name for marker_name, marker_info in self.database_pkl['markers'].items()
+                   if marker_info['clade'] == clade)
+
+
+    def get_all_markers(self):
+        self.load_database()
+        return list(self.database_pkl['markers'].keys())
+
+    def get_markers2ext(self):
+        self.load_database()
+        return {marker_name: ['t__' + sgb for sgb in marker_info['ext']] for marker_name, marker_info in self.database_pkl['markers'].items()}
+
 
     def get_filtered_markers(self, clades):
         """Retrieve the markers belonging to a list of clades
@@ -124,9 +159,6 @@ class MetaphlanDatabaseController:
     def resolve_index(self):
         """Resolves the name to the MPA database
 
-        Args:
-            database (str): the name or path of the database
-
         Returns:
             str: the resolved name to the database
         """
@@ -145,7 +177,7 @@ class MetaphlanDatabaseController:
 
     def __init__(self, database, bowtie2db=None):
         self.mpa_script_folder = os.path.dirname(os.path.abspath(__file__))
-        if bowtie2db == None:
+        if bowtie2db is None:
             self.default_db_folder = os.path.join(
                 self.mpa_script_folder, "..", "metaphlan_databases")
             self.default_db_folder = os.environ.get(
