@@ -5,58 +5,67 @@ __author__ = ('Michal Puncochar (michal.puncochar@unitn.it), '
               'Francesco Asnicar (f.asnicar@unitn.it), '
               'Moreno Zolfo (moreno.zolfo@unitn.it), '
               'Francesco Beghini (francesco.beghini@unitn.it)')
-__version__ = '4.0.6'
-__date__ = '1 Mar 2023'
+__version__ = '4.1.2'
+__date__ = '16 Oct 2025'
 
 
 
-import os
-import time
-import pickle
-import tempfile
-import bz2
-import pickletools
 import argparse as ap
+import bz2
+import os
+import subprocess as sp
+import tempfile
+import time
 from collections import Counter
 from shutil import rmtree
 
+import numpy as np
 import pysam
 import scipy.stats as sps
 
 try:
     from .external_exec import samtools_sam_to_bam, samtools_sort_bam_v1, decompress_bz2
-    from .util_fun import info, error
+    from .util_fun import info, error, warning
     from .parallelisation import execute_pool
     from .database_controller import MetaphlanDatabaseController
-    from .consensus_markers import ConsensusMarker
+    from .consensus_markers import ConsensusMarker, ConsensusMarkers
 except ImportError:
     from external_exec import samtools_sam_to_bam, samtools_sort_bam_v1, decompress_bz2
-    from util_fun import info, error
+    from util_fun import info, error, warning
     from parallelisation import execute_pool
     from database_controller import MetaphlanDatabaseController
-    from consensus_markers import ConsensusMarker
+    from consensus_markers import ConsensusMarker, ConsensusMarkers
 
 
 class SampleToMarkers:
     """SampleToMarkers class"""
 
     class DEFAULTS:
-        pileup_stepper = 'nofilter'
-        poly_error_rate = 0.001
+        min_reads_aligning = 8
+        depth_avg_q = 0.2
+        quasi_marker_frac = 0.33
+        min_breadth = 80
         min_base_coverage = 1
         min_base_quality = 30
         min_mapping_quality = 10
         poly_dominant_frq_thrsh = 0.8
+
+    class CONSTANTS:
+        """Settings not modifiable through arguments"""
+        allowed_bases = set(list('ACTGactg'))
+        pileup_stepper = 'nofilter'
+        poly_error_rate = 0.001
         poly_pvalue_threshold = 0.05
 
+
     def decompress_from_bz2(self):
         """ Decompressed SAM.BZ2 files
 
         Returns:
             (list, str): tuple with the list of decompressed files and their format
         """
-        results = execute_pool(
-            ((SampleToMarkers.decompress_bz2_file, i, self.tmp_dir) for i in self.input), self.nprocs)
+        results = execute_pool(((SampleToMarkers.decompress_bz2_file, i, self.tmp_dir)
+                                for i in self.input), self.nprocs)
         decompressed = [r[0] for r in results]
         decompressed_format = [r[1] for r in results]
         if decompressed_format[1:] == decompressed_format[:-1]:
@@ -91,17 +100,14 @@ class SampleToMarkers:
             info("\tDone.")
         if self.input_format.lower() == "sam":
             info("\tConverting samples to BAM format...")
-            self.input = execute_pool(
-                ((samtools_sam_to_bam, i, self.tmp_dir) for i in self.input), self.nprocs)
+            self.input = execute_pool(((samtools_sam_to_bam, i, self.tmp_dir) for i in self.input), self.nprocs)
             info("\tDone.")
             info("\tSorting BAM samples...")
-            self.input = execute_pool(
-                ((samtools_sort_bam_v1, i, self.tmp_dir) for i in self.input), self.nprocs)
+            self.input = execute_pool(((samtools_sort_bam_v1, i, self.tmp_dir) for i in self.input), self.nprocs)
             info("\tDone.")
         elif not self.sorted:
             info("\tSorting BAM samples...")
-            self.input = execute_pool(
-                ((samtools_sort_bam_v1, i, self.tmp_dir) for i in self.input), self.nprocs)
+            self.input = execute_pool(((samtools_sort_bam_v1, i, self.tmp_dir) for i in self.input), self.nprocs)
             info("\tDone.")
 
         info('\tIndexing BAM samples...')
@@ -116,44 +122,56 @@ class SampleToMarkers:
         """
         for i in self.input:
             info("\tProcessing sample: {}".format(i))
-            results = self.get_consensuses_for_sample(i)
-            self.write_results_as_pkl(filtered, os.path.splitext(
-                os.path.basename(i))[0], results)
+            consensuses, coverages = self.get_consensuses_for_sample(i)
+            consensuses_filtered = self.filter_consensuses(consensuses, coverages)
+            if len(consensuses_filtered) == 0:
+                warning(f'\t\tSkipping sample as it contains no markers after filtering')
+
+            consensus_markers = ConsensusMarkers(consensuses_filtered, self.database_controller.get_database_name())
+            output_filename = f'{os.path.splitext(os.path.basename(i))[0]}{filtered}.json.bz2'
+            output_path = os.path.join(self.output_dir, output_filename)
+            consensus_markers.to_json(output_path)
             info("\tDone.")
 
 
-    def get_consensuses_for_sample(self, input_bam, stepper=DEFAULTS.pileup_stepper,
-                                   error_rate=DEFAULTS.poly_error_rate,
-                                   poly_pvalue_threshold=DEFAULTS.poly_pvalue_threshold):
+    def get_consensuses_for_sample(self, input_bam):
         """Pileup on the specified BAM file
 
         Args:
             input_bam (str): the path to the input BAM file
-            stepper (str): stepper to use in the pileup
-            error_rate (float): the sequencing error rate
-            poly_pvalue_threshold (float): the p-value threshold to call polymorphic position
 
         Returns:
-            list[tuple[str, str]]: the list with the marker names and consensus sequences
+            tuple[dict[str, str], dict[str, np.ndarray[int]]]: the list with the marker names and consensus sequences
         """
 
+        stepper = SampleToMarkers.CONSTANTS.pileup_stepper
+        error_rate = SampleToMarkers.CONSTANTS.poly_error_rate
+        poly_pvalue_threshold = SampleToMarkers.CONSTANTS.poly_pvalue_threshold
+
         info('\t\tLoading the bam file and extracting information...')
-        sam_file = pysam.AlignmentFile(input_bam)
+        sam_file = pysam.AlignmentFile(input_bam, check_sq=False)
+        if not sam_file.references:
+            warning(f'\t\t\tSample {input_bam} has empty BAM file, skipping')
+            return {}, {}
+
 
         all_markers = sam_file.references
         marker_lengths = sam_file.lengths
         marker_to_length = dict(zip(all_markers, marker_lengths))
         info('\t\tDone.')
 
-        ACTGactg = set(list('ACTGactg'))
-
         info('\t\tRunning the pileup...')
         consensuses = {m: bytearray(b'-' * marker_to_length[m]) for m in all_markers}
+        coverages = {m: np.zeros(marker_to_length[m], dtype=int) for m in all_markers}
         for base_pileup in sam_file.pileup(contig=None, stepper=stepper, min_base_quality=self.min_base_quality):
             marker = base_pileup.reference_name
             pos = base_pileup.pos
-            bases = [b.upper() for b in base_pileup.get_query_sequences() if b in ACTGactg]
+            bases = [b.upper() for b in base_pileup.get_query_sequences()
+                     if b in SampleToMarkers.CONSTANTS.allowed_bases]
             base_coverage = len(bases)
+
+            coverages[marker][pos] = base_coverage
+
             if base_coverage < self.min_base_coverage:
                 continue
 
@@ -171,30 +189,60 @@ class SampleToMarkers:
 
         info('\t\tDone.')
 
-        consensuses = [(m, c.decode()) for m, c in consensuses.items()]  # convert bytearrays to strings
-        return consensuses
+        consensuses = {m: c.decode() for m, c in consensuses.items()}  # convert bytearrays to strings
+        return consensuses, coverages
 
 
-
-    def write_results_as_pkl(self, filtered, name, results):
-        """Writes the consensus sequences as a PKL file
+    def filter_consensuses(self, consensuses, coverages):
+        """
+        Filters the markers for colliding quasi-markers and the breadth of coverage.
+        Also computes the depth of coverage
 
         Args:
-            filtered (str): string to append when filtered for clades
-            name (str): the name of the sample
-            results (list): the list with the reconstructed consensus markers
-        """
-        consensus = []
-        for m, seq in results:
-            marker = ConsensusMarker(m, seq)
-            if marker.breadth >= self.breadth_threshold:
-                consensus.append(
-                    {"marker": marker.name, "breath": marker.breadth, "sequence": marker.sequence})
-        with open(os.path.join(self.output_dir, '{}{}.pkl'.format(name, filtered)), 'wb') as markers_pkl:
-            markers_pkl.write(pickletools.optimize(
-                pickle.dumps(consensus, pickle.HIGHEST_PROTOCOL)))
+            consensuses (dict[str, str]): dictionary marker name => sequence
+            coverages: (dict[str, np.ndarray[int]]): dictionary marker name => per-base coverages
 
-    def parallel_filter_sam(self, input_file, filtered_markers):
+        Returns:
+            list[ConsensusMarker]:
+        """
+        markers2ext = self.database_controller.get_markers2ext()
+        markers2clade = self.database_controller.get_markers2clade()
+        clade2nmarkers = Counter(markers2clade.values())
+
+        markers_all = list(consensuses.keys())
+        clades_all = [markers2clade[m] for m in markers_all]
+        clades_counter = Counter(clades_all)  # number of marker with non-zero coverage for each clade
+
+        markers_non_quasi = [m for m in markers_all if len(markers2ext[m]) == 0]
+        markers_quasi = [m for m in markers_all if len(markers2ext[m]) > 0]
+        # if any external SGB has >= quasi_marker_frac non-zero markers ==> discard
+        markers_quasi_filtered = [m for m in markers_quasi
+                                  if not any(clades_counter[ext_sgb] / clade2nmarkers[ext_sgb] >= self.quasi_marker_frac
+                                             for ext_sgb in markers2ext[m])]
+
+        consensus_markers_filtered = []
+        for m in markers_non_quasi + markers_quasi_filtered:
+            seq = consensuses[m]
+            cov = coverages[m]
+            assert len(seq) == len(cov)
+
+            # calculate robust average by trimming upper and lower quantiles at depth_avg_q
+            cov = sorted(cov)
+            t = int(self.depth_avg_q * len(cov))
+            cov_trimmed = cov[t:-t]
+            avg_depth = np.mean(cov_trimmed)
+
+            c = ConsensusMarker(m, seq, avg_depth=avg_depth)
+            if c.breadth < self.breadth_threshold:
+                continue
+            consensus_markers_filtered.append(c)
+
+        return consensus_markers_filtered
+
+
+    @staticmethod
+    def parallel_filter_sam(input_file, tmp_dir, input_format, min_mapping_quality, min_reads_aligning,
+                            filtered_markers, all_markers):
         """
         Filters an input SAM file
             * filters out viral markers (VDB)
@@ -203,52 +251,60 @@ class SampleToMarkers:
 
         Args:
             input_file (str): the input SAM file
+            tmp_dir:
+            input_format:
+            min_reads_aligning:
+            min_mapping_quality:
             filtered_markers (set): the list with the markers of the filtered clades, None if to use all markers
+            all_markers:
 
         Returns:
             str: the path to the output file
         """
-        output_file = os.path.join(self.tmp_dir, input_file.split('/')[-1])
-        if self.input_format.lower() == "bz2":
-            ifn = bz2.open(input_file, 'rt')
+        output_file = os.path.join(tmp_dir, input_file.split('/')[-1])
+        if input_format.lower() == "bz2":
+            ifn = bz2.open(input_file, 'rb')
             output_file = output_file.replace('.bz2', '')
         else:
-            assert self.input_format.lower() == "sam"
-            ifn = open(input_file, 'r')
+            assert input_format.lower() == "sam"
+            ifn = open(input_file, 'rb')
 
         def filter_mapping_line(line_fields_, markers_subset):
             marker_ = line_fields_[2]
-            if marker_.startswith('VDB'):
+            if marker_.startswith(b'VDB'):
                 return False
-            if markers_subset is not None and marker_ not in markers_subset:
+            if markers_subset is not None and marker_.decode() not in markers_subset:
                 return False
-            mapq_ = int(line_fields_[4])
-            if mapq_ < self.min_mapping_quality:
+            mapq_ = int(line_fields_[4].decode())
+            if mapq_ < min_mapping_quality:
                 return False
             return True
 
         marker_to_reads = Counter()
         for line in ifn:
-            if line.startswith('@'):
+            if line.startswith(b'@'):
                 continue
-            line_fields = line.rstrip('\n').split('\t')
+            line_fields = line.rstrip(b'\n').split(b'\t')
             if filter_mapping_line(line_fields, filtered_markers):
-                marker = line_fields[2]
+                marker = line_fields[2].decode()
                 marker_to_reads[marker] += 1
+                if marker not in all_markers:
+                    error(f'Marker {marker} not in the metaphlan database', exit=True)
 
-        selected_markers = set((m for m, c in marker_to_reads.items() if c >= self.min_reads_aligning))
+        selected_markers = set((m for m, c in marker_to_reads.items() if c >= min_reads_aligning))
 
         ifn.seek(0)  # second pass of the file
-        with open(output_file, 'w') as ofn:
+        with open(output_file, 'wb') as ofn:
             for line in ifn:
-                line_fields = line.rstrip('\n').split('\t')
-                if line.startswith('@HD'):
-                    ofn.write(line)
-                elif line.startswith('@'):
-                    marker = line_fields[1].split(':')[1]
+                line_fields = line.rstrip(b'\n').split(b'\t')
+                if line.startswith(b'@SQ'):
+                    assert line_fields[1].startswith(b'SN:')  # in bowtie2 output SN is always the first tag
+                    marker = line_fields[1][3:].decode()
                     if marker in selected_markers:
                         ofn.write(line)
-                else:
+                elif line.startswith(b'@'):  # other header lines like @HD and @PG
+                    ofn.write(line)
+                else:  # mapping lines
                     if filter_mapping_line(line_fields, selected_markers):
                         ofn.write(line)
 
@@ -256,11 +312,12 @@ class SampleToMarkers:
         return output_file
 
     def filter_sam_files(self):
-        """Filters the input SAM files with the hits against markers of specific clades"""
-        filtered_markers = self.database_controller.get_filtered_markers(
-            self.clades) if len(self.clades) > 0 else None
-        self.input = execute_pool(
-            ((self.parallel_filter_sam, i, filtered_markers) for i in self.input), self.nprocs)
+        """Filters the input SAM files with the hits against markers of specific clades and low quality reads"""
+        filtered_markers = self.database_controller.get_filtered_markers(self.clades) if len(self.clades) > 0 else None
+        all_markers = set(self.database_controller.get_all_markers())
+        self.input = execute_pool(((SampleToMarkers.parallel_filter_sam, i, self.tmp_dir, self.input_format,
+                                    self.min_mapping_quality, self.min_reads_aligning, filtered_markers, all_markers)
+                                   for i in self.input), self.nprocs)
         self.input_format = 'sam'
         self.sorted = False
 
@@ -277,28 +334,30 @@ class SampleToMarkers:
         self.convert_inputs()
         info("Done.")
         info("Getting consensus markers from samples...")
-        self.build_consensus_markers(
-            filtered='_filtered' if len(self.clades) > 0 else '')
+        self.build_consensus_markers(filtered='_filtered' if len(self.clades) > 0 else '')
         info("Done.")
         if not self.debug:
             info("Removing temporary files...")
             rmtree(self.tmp_dir)
             info("Done.")
 
+
     def __init__(self, args):
         self.input = args.input
-        self.sorted = args.sorted
-        self.input_format = args.input_format
         self.output_dir = args.output_dir
         self.database_controller = MetaphlanDatabaseController(args.database)
+        self.tmp_dir = args.output_dir if args.tmp is None else args.tmp
         self.breadth_threshold = args.breadth_threshold
+        self.input_format = args.input_format
+        self.clades = args.clades
+        self.sorted = args.sorted
         self.min_reads_aligning = args.min_reads_aligning
         self.min_base_coverage = args.min_base_coverage
         self.min_base_quality = args.min_base_quality
         self.min_mapping_quality = args.min_mapping_quality
         self.dominant_frq_threshold = args.dominant_frq_threshold
-        self.clades = args.clades
-        self.tmp_dir = args.output_dir if args.tmp is None else args.tmp
+        self.quasi_marker_frac = args.quasi_marker_frac
+        self.depth_avg_q = args.depth_avg_q
         self.debug = args.debug
         self.nprocs = args.nprocs
 
@@ -309,22 +368,19 @@ def read_params():
     Returns:
         namespace: The populated namespace with the command line arguments
     """
-    p = ap.ArgumentParser(
-        description="", formatter_class=ap.ArgumentDefaultsHelpFormatter)
-    p.add_argument('-i', '--input', type=str,
-                   nargs='+', default=[],
-                   help="The input samples as SAM or BAM files")
-    p.add_argument('--sorted', action='store_true', default=False,
-                   help="Whether the BAM input files are sorted")
-    p.add_argument('-f', '--input_format', type=str, default="bz2",
-                   help="The input samples format {bam, sam, bz2}")
-    p.add_argument('-o', '--output_dir', type=str, default=None,
-                   help="The output directory")
+    p = ap.ArgumentParser(description="", formatter_class=ap.ArgumentDefaultsHelpFormatter)
+    # required
+    p.add_argument('-i', '--input', type=str, nargs='+', required=True, help="The input samples as SAM or BAM files")
+    p.add_argument('-o', '--output_dir', type=str, required=True, help="The output directory")
+
+    # optional
     p.add_argument('-d', '--database', type=str, default='latest',
-                   help="The input MetaPhlAn " + __version__ + " database")
-    p.add_argument('-b', '--breadth_threshold', type=int, default=80,
-                   help="The breadth of coverage threshold for the consensus markers")
-    p.add_argument('--min_reads_aligning', type=int, default=8,
+                   help="The input MetaPhlAn " + __version__ + " database (path to the pkl file)")
+    p.add_argument('--clades', type=str, nargs='+', default=[],
+                   help="Restricts the reconstruction of the markers to the specified clades")
+    p.add_argument('-f', '--input_format', type=str, default="bz2", help="The input samples format {bam, sam, bz2}")
+    p.add_argument('--sorted', action='store_true', default=False, help="Whether the BAM input files are sorted")
+    p.add_argument('--min_reads_aligning', type=int, default=SampleToMarkers.DEFAULTS.min_reads_aligning,
                    help="The minimum number of reads to cover a marker")
     p.add_argument('--min_base_coverage', type=int, default=SampleToMarkers.DEFAULTS.min_base_coverage,
                    help="The minimum depth of coverage for a base to be considered")
@@ -334,15 +390,20 @@ def read_params():
                    help="The minimum quality for a mapping of the read to be considered.")
     p.add_argument('--dominant_frq_threshold', type=float, default=SampleToMarkers.DEFAULTS.poly_dominant_frq_thrsh,
                    help="The cutoff for degree of 'allele dominance' for a position to be considered polymorphic")
-    p.add_argument('--clades', type=str, nargs='+', default=[],
-                   help="Restricts the reconstruction of the markers to the specified clades")
+    p.add_argument('--quasi_marker_frac', type=float, default=SampleToMarkers.DEFAULTS.quasi_marker_frac,
+                   help="Fraction [0-1] of markers with a hit of an external SGB to disqualify a quasi-marker.")
+    p.add_argument('--depth_avg_q', type=float, default=SampleToMarkers.DEFAULTS.depth_avg_q,
+                   help="A quantile to cut from both ends of the coverage distributions to calculate robust average.")
     p.add_argument('--tmp', type=str, default=None,
-                   help="If specified, the directory where to store the temporal files")
+                   help="If specified, the directory where to store the temporary files. "
+                        "Otherwise the output directory will be used.")
+    p.add_argument('-b', '--breadth_threshold', type=int, default=SampleToMarkers.DEFAULTS.min_breadth,
+                   help="The breadth of coverage threshold for the consensus markers")
     p.add_argument('--debug', action='store_true', default=False,
-                   help="If specified, StrainPhlAn will not remove the temporal folder. "
+                   help="If specified, StrainPhlAn will not remove the temporary folder. "
                         "Not available with inputs in BAM format")
-    p.add_argument('-n', '--nprocs', type=int, default=1,
-                   help="The number of threads to execute the script")
+    p.add_argument('-n', '--nprocs', type=int, default=1, help="The number of threads to execute the script")
+
     return p.parse_args()
 
 
@@ -352,25 +413,19 @@ def check_params(args):
     Args:
         args (namespace): the arguments to check
     """
-    if not args.input:
-        error('-i (or --input) must be specified', exit=True)
-    elif not args.input_format:
-        error('-f (or --input_format) must be specified', exit=True)
-    elif not args.output_dir:
-        error('-o (or --output_dir) must be specified', exit=True)
-    elif args.input_format.lower() not in ['bam', 'sam', 'bz2']:
-        error('The input format must be SAM, BAM, or compressed in BZ2 format', exit=True)
-    elif args.input_format.lower() == "bam" and len(args.clades) > 0:
-        error('The --clades option cannot be used with inputs in BAM format', exit=True)
-    elif not os.path.exists(args.output_dir):
-        error('The directory {} does not exist'.format(
-            args.output_dir), exit=True)
-    elif not (args.tmp is None) and not os.path.exists(args.tmp):
+    if not os.path.exists(args.output_dir):
+        error('The directory {} does not exist'.format(args.output_dir), exit=True)
+    if args.tmp is not None and not os.path.exists(args.tmp):
         error('The directory {} does not exist'.format(args.tmp), exit=True)
-    elif args.database != 'latest' and not os.path.exists(args.database):
+    if args.database != 'latest' and not os.path.exists(args.database):
         error('The database does not exist', exit=True)
-    else:
-        check_input_files(args.input, args.input_format)
+    if args.input_format.lower() not in ['bam', 'sam', 'bz2']:
+        error('The input format must be SAM, BAM, or compressed in BZ2 format', exit=True)
+    if args.input_format.lower() == "bam" and len(args.clades) > 0:
+        error('The --clades option cannot be used with inputs in BAM format', exit=True)
+
+    check_input_files(args.input, args.input_format)
+
     return args
 
 
@@ -390,16 +445,34 @@ def check_input_files(input_files, input_format):
                 s, input_format.upper()), exit=True)
 
 
+def check_samtools():
+    samtools_version_output = sp.run("samtools", capture_output=True).stderr.decode().split('\n')
+    for line in samtools_version_output:
+        if line.startswith('Version:'):
+            samtools_version_info = line
+            break
+    else:
+        error("Couldn't get information about the samtools version", exit=True)
+        return
+
+    samtools_version = samtools_version_info.rstrip('\n').split(' ')[1]
+    samtools_version_parts = [int(x) for x in samtools_version.split('.')]
+    if samtools_version_parts < [1]:
+        error(f'The samtools version required is >= 1.x.x, installed is {samtools_version}', exit=True)
+    info(f'Using samtools version {samtools_version}')
+
+
 def main():
     t0 = time.time()
     args = read_params()
     info("Start samples to markers execution")
+    check_samtools()
     check_params(args)
     sampletomarkers = SampleToMarkers(args)
     sampletomarkers.run_sample2markers()
     exec_time = time.time() - t0
-    info("Finish samples to markers execution ({} seconds): Results are stored at \"{}\"".format(
-        round(exec_time, 2), args.output_dir))
+    info("Finish samples to markers execution ({} seconds): Results are stored at "
+         "\"{}\"".format(round(exec_time, 2), args.output_dir))
 
 
 if __name__ == '__main__':
