@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 __author__ = 'Aitor Blanco (aitor.blancomiguez@unitn.it'
-__version__ = '4.0.6'
-__date__ = '1 Mar 2023'
+__version__ = '4.1.2'
+__date__ = '16 Oct 2025'
 
 import os
 import time
@@ -70,7 +70,7 @@ def get_gtdb_profile(mpa_profile, gtdb_profile, database):
                 elif 't__SGB' in line:
                     line = line.strip().split('\t')
                     gtdb_tax = sgb2gtdb[line[0].split('|')[-1][3:]]
-                    if gtdb_tax not in abundances:
+                    if gtdb_tax not in abundances['s']:
                         abundances['s'][gtdb_tax] = 0
                     abundances['s'][gtdb_tax] += float(line[2])
         tax_levels.reverse()
@@ -80,11 +80,12 @@ def get_gtdb_profile(mpa_profile, gtdb_profile, database):
                 new_level = tax_levels[tax_levels.index(tax_level)+1]
                 if new_tax not in abundances[new_level]:
                     abundances[new_level][new_tax] = 0
-                abundances[new_level][new_tax] += abundances[tax_level][tax]
+                abundances[new_level][new_tax] += abundances[tax_level][tax]        
         tax_levels.reverse()
+        total4level = {x:sum([y for x,y in abundances[x].items()]) for x in tax_levels}
         for tax_level in tax_levels:
             for tax in abundances[tax_level]:
-                wf.write('{}\t{}\n'.format(tax, abundances[tax_level][tax]))
+                wf.write('{}\t{}\n'.format(tax, abundances[tax_level][tax] * 100 / total4level[tax_level]))
 
 
 def main():
