@@ -3,10 +3,11 @@ __author__ = ('Aitor Blanco Miguez (aitor.blancomiguez@unitn.it), '
               'Francesco Asnicar (f.asnicar@unitn.it), '
               'Moreno Zolfo (moreno.zolfo@unitn.it), '
               'Francesco Beghini (francesco.beghini@unitn.it)')
-__version__ = '4.0.6'
-__date__ = '1 Mar 2023'
-
+__version__ = '4.1.2'
+__date__ = '16 Oct 2025'
 
+import bz2
+import gzip
 import os
 import sys
 import time
@@ -64,7 +65,7 @@ def error(message, init_new_line=True, exit=False, exit_value=1):
         sys.stdout.write('\n')
 
     if exit:
-        sys.stderr.write('{}: Stop StrainPhlAn execution.\n'.format(
+        sys.stderr.write('{}: Stop execution.\n'.format(
             time.ctime(int(time.time()))))
         sys.exit(exit_value)
 
@@ -82,3 +83,13 @@ def create_folder(path):
             os.mkdir(path)
     except Exception as e:
         error('An error ocurred when creating the \"{}\" folder'.format(e), exit=True)
+
+
+def openrt(file_path):
+    file_path = str(file_path)
+    if file_path.endswith('.bz2'):
+        return bz2.open(file_path, 'rt')
+    elif file_path.endswith('.gz'):
+        return gzip.open(file_path, 'rt')
+    else:
+        return open(file_path)
